from hashlib import md5

from fastapi import FastAPI, Request

from popol.utils import get_settings


def get_cache_key(request: Request) -> str:
    """
    Generate cache keys using the concatenation of route paths, HTTP methods and query params.
    And also available `CACHE_INCLUDE_IP_ADDRESS` configuration to insert client ip address to key.

    Returns:
        str: The cache key (md5 hash).
    """

    path = request.url.path
    query_params = request.scope.get("query_string", b"").decode()
    part = path + request.method + query_params
    app: FastAPI = request.app
    settings = get_settings(app)
    if getattr(settings, "CACHE_INCLUDE_IP_ADDRESS", True):
        ip_address = request.client.host
        if ip_address:
            part += ip_address

    key = md5(part.encode()).hexdigest()
    return key



from ..serializers.base import BaseSerializer
from .base import BaseCacheBackend


class DummyBackend(BaseCacheBackend):
    def __init__(self, serializer: BaseSerializer, **options):
        super().__init__(serializer, **options)

    def connect(self) -> "DummyBackend":
        pass

    def disconnect(self) -> None:
        pass

    def get(self, key):
        pass

    def set(self, key, value, timeout=None):
        pass

    def delete(self, key):
        pass

    def clear(self):
        pass



import pickle
from typing import Any, Optional, Union

from .base import BaseSerializer


class PickleSerializer(BaseSerializer):
    """
    Pickle serializer.
    """

    def __init__(
        self, dumps_kwargs: Optional[dict] = None, loads_kwargs: Optional[dict] = None
    ):
        self.dumps_kwargs = dumps_kwargs or {}
        self.loads_kwargs = loads_kwargs or {}
        self.dumps_kwargs.setdefault("protocol", pickle.HIGHEST_PROTOCOL)

    def dumps(self, data: Any) -> bytes:
        if isinstance(data, int):
            return data

        return pickle.dumps(data, **self.dumps_kwargs)

    def loads(self, data: Union[bytes, int]) -> Any:
        try:
            return int(data)
        except ValueError:
            return pickle.loads(data, **self.loads_kwargs)



import os
from asyncio import iscoroutinefunction
from email.encoders import encode_base64
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate, make_msgid
from mimetypes import guess_type
from typing import List, Optional, TextIO

from pydantic import EmailStr, validate_arguments

from ..utils import run_sync


class Message:
    @validate_arguments
    def __init__(
        self,
        *,
        from_email: str,
        subject: str,
        recipients: List[EmailStr],
        text: Optional[str] = None,
        html: Optional[str] = None,
        cc: List[EmailStr] = [],
        bcc: List[EmailStr] = [],
        reply_to: List[EmailStr] = [],
    ) -> None:
        self.subject = subject
        self.recipients = recipients
        self.text = text
        self.html = html
        self.cc = cc
        self.bcc = bcc
        self.reply_to = reply_to
        self.from_email = from_email
        self.id = make_msgid()
        self._message: MIMEMultipart = MIMEMultipart("alternative")

    def attach_file(
        self,
        file: TextIO,
        *,
        filename: str = None,
        mimetype: str = None,
        headers: dict = {},
    ) -> None:
        """Attach a file to the message.

        The file will be attached with the default MIME type application/octet-stream.
        """

        filename = filename or os.path.basename(file.name)
        mimetype = mimetype or guess_type(filename)[0] or "application/octet-stream"
        main_type, sub_type = mimetype.split("/", 1)
        part = MIMEBase(main_type, sub_type)
        if iscoroutinefunction(file.read):
            payload = run_sync(file.read)
        else:
            payload = file.read()
        part.set_payload(payload)
        encode_base64(part)
        part.add_header("Content-Disposition", "attachment", filename=filename)
        for k, v in headers.items():
            part.add_header(k, v)
        self._message.attach(part)

    def build(self) -> MIMEMultipart:
        self._message["From"] = self.from_email
        self._message["Subject"] = self.subject
        self._message["To"] = ", ".join(self.recipients)

        if self.cc:
            self._message["Cc"] = ", ".join(self.cc)
        if self.bcc:
            self._message["Bcc"] = ", ".join(self.bcc)
        if self.reply_to:
            self._message["Reply-To"] = ", ".join(self.reply_to)
        if self.text:
            self._message.attach(MIMEText(self.text, "plain"))
        if self.html:
            self._message.attach(MIMEText(self.html, "html"))

        self._message["Message-ID"] = self.id
        self._message["Date"] = formatdate(localtime=True)
        return self._message



from abc import ABCMeta, abstractmethod
from datetime import timedelta
from typing import Any, Dict, Optional, Union

from popol.cache.serializers.base import BaseSerializer

DictStrAny = Dict[str, Any]


class BaseCacheBackend(metaclass=ABCMeta):
    """
    Base class for all cache backends.
    """

    def __init__(self, serializer: BaseSerializer, **kwargs) -> None:
        self.serializer = serializer
        self.kwargs: DictStrAny = kwargs
        self.client = None

    @abstractmethod
    def connect(self) -> "BaseCacheBackend":
        """
        Connect to the server.
        """

    @abstractmethod
    def disconnect(self) -> None:
        """
        Disconnect from the server.
        """

    @abstractmethod
    def get(self, key: Any) -> Any:
        """
        Get a value from the cache.
        """

    @abstractmethod
    def set(
        self, key: Any, value: Any, timeout: Optional[Union[int, timedelta]] = None
    ) -> None:
        """
        Set a value in the cache.
        """

    @abstractmethod
    def delete(self, key: Any) -> None:
        """
        Delete a value from the cache.
        """

    @abstractmethod
    def clear(self) -> None:
        """
        Clear the entire cache.
        """



from typing import Any, Dict, Tuple

from .queue import Queue


def parse_config(settings: Any) -> Tuple[Dict[str, Queue], Dict[str, dict]]:
    """
    SAQ configuration parsing.

    Args:
        settings: The settings (can be pydantic.BaseSettings).

    Returns:
        Tuple[Dict[str, Queue], Dict[str, dict]]: The SAQ queues and the queue settings.
    """

    saq_queues: Dict[str, dict] = getattr(settings, "SAQ_QUEUES", {})
    if not isinstance(saq_queues, dict):
        raise RuntimeError("SAQ_QUEUES must be a dict, got {}".format(type(saq_queues)))

    queue_maps = {}
    queue_settings = {}
    for q_name, q_param in saq_queues.items():
        url = q_param.get("url", None)
        if not url:
            raise RuntimeError("No url specified for queue {}".format(q_name))

        queue = Queue.from_url(url, q_name)
        queue_maps[q_name] = queue
        queue_settings[q_name] = q_param
    return queue_maps, queue_settings



from typing import Any, Dict

from fastapi import FastAPI

from ...utils import get_settings
from .config import parse_config
from .queue import Queue


def setup(app: FastAPI, settings: Any = None) -> Dict[str, Queue]:
    """
    Install the saq plugin to the app.
    This will install 2 new attributes to `app.state` which are:

    * `saq_queue` - SAQ Queue (Default, if available)
    * `saq_queues` - All SAQ queues (dict type)

    Args:
        app: FastAPI app.
        settings: The settings (can be pydantic.BaseSettings).

    Returns:
        Dict[str, Queue]: The SAQ queues.
    """

    settings = get_settings(app, settings)
    queue_maps, _ = parse_config(settings)
    app.state.saq_queue = queue_maps.get("default")
    app.state.saq_queues = queue_maps
    return queue_maps



import os
import sys
from distutils.dir_util import copy_tree
from importlib import import_module
from typing import Any, Coroutine, TypeVar

from asyncer import syncify
from fastapi import FastAPI, HTTPException

T = TypeVar("T")
POPOL_TEMPLATE_DIR = os.path.join(os.path.dirname(__file__), "templates")


def copy_template_dir(dest_dir: str, *, project_name: str, project_description: str):
    """
    Copy the template directory to the destination directory.

    Args:
        dest_dir (str): The destination directory.
        project_name (str): The project name.
        project_description (str): The project description.
    """

    dest_dir = os.path.abspath(dest_dir)
    os.makedirs(dest_dir, exist_ok=True)
    copy_tree(POPOL_TEMPLATE_DIR, dest_dir)
    readme_path = os.path.join(dest_dir, "README.md")
    with open(readme_path, "r") as f:
        old_data = f.read()
        new_data = old_data.format(
            project_name=project_name, project_description=project_description
        )

    with open(readme_path, "w") as f:
        f.write(new_data)


def import_attr(module: str) -> Any:
    """
    Import attributes from a module.

    Args:
        module: Module name (e.g. "os.path")

    Returns:
        Any: Imported attributes
    """

    package, attr = module.rsplit(".", 1)
    mod = import_module(package)
    return getattr(mod, attr)


def get_settings(app: FastAPI, settings: T = None) -> T:
    """
    Get the settings from the application.

    Args:
        app: The application.
        settings: The settings.

    Raises:
        RuntimeError: If settings not found.

    Returns:
        Any: The settings.
    """

    if not hasattr(app, "state"):
        raise RuntimeError("The application does not have a state.")

    settings = getattr(app.state, "settings", settings)
    if not settings:
        raise RuntimeError("The application does not have a settings.")

    return settings


_current_app = None


def load_app() -> FastAPI:
    """
    Load the application.

    Returns:
        FastAPI: The application.
    """

    global _current_app

    if _current_app and isinstance(_current_app, FastAPI):
        return _current_app

    cwd = os.getcwd()
    sys.path.insert(0, cwd)
    env = os.getenv("POPOL_APP", "app.main:app")
    try:
        module, attr = env.rsplit(":", 1)
    except ValueError:
        raise RuntimeError("Invalid POPOL_APP environment variable.")

    try:
        _current_app = import_attr(f"{module}.{attr}")
    except ModuleNotFoundError:
        raise RuntimeError(
            f"Could not import {module} (make sure you are in the app's root directory)"
        )

    assert isinstance(_current_app, FastAPI)
    return _current_app


def abort(status_code: int, detail: Any = None, headers: dict = None):
    """
    Abort the request.

    Args:
        status_code: The status code.
        detail: The detail.
        headers: The headers.

    Raises:
        HTTPException: The HTTP exception.
    """

    raise HTTPException(status_code=status_code, detail=detail, headers=headers)


def run_sync(func: Coroutine, *args, **kwds):
    """
    Run async function to sync.

    Args:
        func: coroutine function
        \*args: argument to pass to func
        \*\*kwds: keyword argument to pass to func
    """

    return syncify(func)(*args, **kwds)



from fastapi import FastAPI

from app.extensions import init_extensions
from app.middleware import init_middleware
from app.routers import init_routers
from app.settings import init_settings

app = FastAPI()
init_settings(app)
init_extensions(app)
init_middleware(app)
init_routers(app)



from fastapi import FastAPI
from pydantic import BaseSettings

from ..utils import get_settings, import_attr
from .backend import EmailBackend


def setup(app: FastAPI, settings: BaseSettings = None) -> EmailBackend:
    """
    Install the email plugin to the app.
    This will attach 1 attribute to `app.state` i.e:

    * `email` - `popol.email.backend.EmailBackend` instance object for sending email.

    Args:
        app: FastAPI app.
        settings: The settings (can be pydantic.BaseSettings).

    Returns:
        popol.email.backend.EmailBackend: The email backend.
    """

    settings = get_settings(app, settings)
    params = {}
    prefix = "EMAIL_"
    email_backend_setting = prefix + "BACKEND"
    for setting_name in dir(settings):
        if not setting_name.startswith(prefix) or setting_name == email_backend_setting:
            continue

        value = getattr(settings, setting_name)
        setting_name = setting_name[len(prefix) :].lower()
        params[setting_name] = value

    klass = getattr(settings, email_backend_setting)
    try:
        email_backend_class = import_attr(klass)
    except Exception as e:
        raise RuntimeError(f"Unable to import email backend {klass}") from e

    email_backend: EmailBackend = email_backend_class(**params)
    app.state.email = email_backend
    return email_backend



from typing import Union

from fastapi import FastAPI
from pydantic import BaseSettings

from ...utils import get_settings

try:
    from sqlalchemy.engine import Engine
    from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
    from sqlmodel import Session, SQLModel, create_engine
    from sqlmodel.ext.asyncio.session import AsyncSession

except ImportError:
    raise RuntimeError(
        "SQLModel is not installed. Please install it with `pip install sqlmodel pyhumps`"
    )


class Database:
    """
    A class to wrap the sqlalchemy engine and open a connection session to the db.
    """

    def __init__(self, engine: Union[Engine, AsyncEngine], is_async: bool = False):
        self.engine = engine
        self.is_async = is_async

    def open(self) -> Union[Session, AsyncSession]:
        if self.is_async:
            return AsyncSession(self.engine)
        else:
            return Session(self.engine)


def setup(app: FastAPI, settings: BaseSettings = None) -> Database:
    """
    Install the sqlmodel plugin to the app.
    This will attach 1 attribute to `app.state` i.e:

    * `db` - `popol.sqlmodel.Database` instance object to open db connection.

    Args:
        app: FastAPI app.
        settings: The settings (can be pydantic.BaseSettings).

    Returns:
        Database: The database.
    """

    settings = get_settings(app, settings)
    prefix = "SQLALCHEMY_"
    db_uri = getattr(settings, f"{prefix}DATABASE_URI", None)
    if not db_uri:
        raise RuntimeError(f"{prefix}DATABASE_URI is not set")

    async_mode = getattr(settings, f"{prefix}ASYNC_MODE", False)
    options = getattr(settings, f"{prefix}OPTIONS", {})
    if async_mode:
        engine = create_async_engine(db_uri, **options)
    else:
        engine = create_engine(db_uri, **options)

    db = Database(engine, async_mode)
    app.state.db = db

    async def startup():
        # reference: https://github.com/tiangolo/sqlmodel/issues/54#issue-981884262
        if async_mode:
            async with engine.begin() as conn:
                await conn.run_sync(SQLModel.metadata.create_all)
        else:
            SQLModel.metadata.create_all(engine)

    app.add_event_handler("startup", startup)
    return db



from typing import Dict

from popol.context.globals import state
from popol.context.proxy import LocalProxy

from .backends.base import BaseCacheBackend

cache: BaseCacheBackend = LocalProxy(lambda: state.cache)
caches: Dict[str, BaseCacheBackend] = LocalProxy(lambda: state.caches)



"""
Popol is a library that provides as-is tools for use on FastAPI.
These are the features currently available:

* Caching - Provides a simple API for saving results to a temporary database.
* ORM Integration
    * SQLModel - Provides a plugin to seamlessly integrate SQLModel with FastAPI.
* Email - Provides a simple API for sending emails.
* Background Jobs:
    * SAQ - SAQ queue integration support for task scheduling.
* Pagination

"""



from inspect import iscoroutinefunction
from typing import Any, Dict, List

from fastapi import FastAPI

from ..utils import get_settings, import_attr
from .backends.base import BaseCacheBackend
from .backends.dummy import DummyBackend
from .serializers.base import BaseSerializer


def setup(app: FastAPI, settings: Any = None) -> Dict[str, BaseCacheBackend]:
    """
    Install the cache plugin to the app.
    This will attach 2 attributes to `app.state` i.e:

    * `cache` - Default cache backend if available.
    * `caches` - All cache backends (dict type)

    Args:
        app: FastAPI app.
        settings: The settings (can be pydantic.BaseSettings).

    Returns:
        Dict[str, popol.cache.backends.base.BaseCacheBackend]: cache backends.
    """
    settings = get_settings(app, settings)
    caches = {}
    caches_settings: Dict[str, dict] = getattr(settings, "CACHES", {})
    for name, cache_config in caches_settings.items():
        serializer_settings: dict = cache_config.get("serializer", {})
        serializer_class = serializer_settings.get(
            "class", "popol.cache.serializers.PickleSerializer"
        )
        serializer_options: dict = serializer_settings.get("options", {})
        try:
            serializer_class: BaseSerializer = import_attr(serializer_class)
        except ImportError as e:
            raise ImportError(
                f"Could not import serializer class {serializer_class}"
            ) from e

        serializer = serializer_class(
            serializer_options.get("dumps", {}),
            serializer_options.get("loads", {}),
        )
        backend = cache_config.get("backend")
        if not backend:
            raise RuntimeError('No backend specified for cache "{}"'.format(name))

        try:
            backend = import_attr(backend)
        except ImportError as e:
            raise RuntimeError(
                'Could not import cache backend "{}"'.format(backend)
            ) from e

        options = cache_config.get("options", {})
        backend = backend(serializer, **options)
        caches[name] = backend

    app.state.caches = caches
    default_cache = caches.get("default")
    if not default_cache:
        default_cache = DummyBackend(None)

    app.state.cache = default_cache

    async def on_shutdown():
        caches: List[BaseCacheBackend] = app.state.caches.values()
        for cache in caches:
            method = cache.disconnect
            if iscoroutinefunction(method):
                await method()
            else:
                method()

    app.add_event_handler("shutdown", on_shutdown)
    return caches



from datetime import datetime

try:
    from humps.main import depascalize
    from sqlalchemy import Column, DateTime
    from sqlalchemy.orm.decl_api import declared_attr
    from sqlmodel import Field, SQLModel

except ImportError:
    raise RuntimeError(
        "SQLModel is not installed. Please install it with `pip install sqlmodel pyhumps`"
    )


class Model(SQLModel):
    """
    Abstract model providing `id`, `date_created` and `date_updated` fields.
    And also automatic table naming to `snake_case`.
    """

    id: int = Field(primary_key=True)
    date_created: datetime = Field(sa_column=Column(DateTime, default=datetime.utcnow))
    date_updated: datetime = Field(sa_column=Column(DateTime, onupdate=datetime.utcnow))

    @declared_attr
    def __tablename__(cls):
        """
        Convert Pascal class name style to `snake_case`
        """
        return depascalize(cls.__name__)



"""
Taken from: https://github.com/aprilahijriyan/falca/blob/main/falca/compat.py
"""

from functools import partial

from pydantic import BaseModel


def _json_default(o: object):
    if isinstance(o, BaseModel):
        return o.dict()
    raise ValueError(f"unknown object {o!r}")


try:
    import orjson as json

    _dumps = json.dumps

    def dumps(*args, **kwds):
        option = kwds.get("option")
        if not option:
            option = json.OPT_NON_STR_KEYS
        kwds["option"] = option
        return _dumps(*args, **kwds)

except ImportError:  # pragma: no cover
    try:
        import rapidjson as json

        dumps = json.dumps

    except ImportError:
        import json

        dumps = json.dumps

dump = getattr(json, "dump", None)
load = getattr(json, "load", None)
loads = json.loads
dumps = partial(dumps, default=_json_default)



"""
Reference: https://github.com/pallets/werkzeug/blob/main/src/werkzeug/local.py
"""

import copy
import math
import operator
import typing as t
from functools import partial

F = t.TypeVar("F", bound=t.Callable[..., t.Any])


class _ProxyLookup:
    """
    Descriptor that handles proxied attribute lookup for
    class `LocalProxy`.
    """

    __slots__ = ("bind_f", "fallback", "is_attr", "class_value", "name")

    def __init__(
        self,
        f: t.Optional[t.Callable] = None,
        fallback: t.Optional[t.Callable] = None,
        class_value: t.Optional[t.Any] = None,
        is_attr: bool = False,
    ) -> None:
        """
        Args:
            f: The built-in function this attribute is accessed through.
            Instead of looking up the special method, the function call is redone on the object.
            fallback: Return this function if the proxy is unbound instead of raising a `RuntimeError`.
            is_attr: This proxied name is an attribute, not a function. Call the fallback immediately to get the value.
            class_value: Value to return when accessed from the ``LocalProxy`` class directly. Used for ``__doc__`` so building docs still works.
        """
        bind_f: t.Optional[t.Callable[["LocalProxy", t.Any], t.Callable]]

        if hasattr(f, "__get__"):
            # A Python function, can be turned into a bound method.

            def bind_f(instance: "LocalProxy", obj: t.Any) -> t.Callable:
                return f.__get__(obj, type(obj))  # pragma: no cover # type: ignore

        elif f is not None:
            # A C function, use partial to bind the first argument.

            def bind_f(instance: "LocalProxy", obj: t.Any) -> t.Callable:
                return partial(f, obj)  # type: ignore

        else:
            # Use getattr, which will produce a bound method.
            bind_f = None

        self.bind_f = bind_f
        self.fallback = fallback
        self.class_value = class_value
        self.is_attr = is_attr

    def __set_name__(self, owner: "LocalProxy", name: str) -> None:
        self.name = name

    def __get__(self, instance: "LocalProxy", owner: t.Optional[type] = None) -> t.Any:
        if instance is None:
            if self.class_value is not None:  # pragma: no cover
                return self.class_value  # pragma: no cover

            return self  # pragma: no cover

        try:
            obj = instance._get_current_object()
        except RuntimeError:
            if self.fallback is None:
                raise

            fallback = self.fallback.__get__(instance, owner)  # type: ignore

            if self.is_attr:
                # __class__ and __doc__ are attributes, not methods.
                # Call the fallback to get the value.
                return fallback()

            return fallback  # pragma: no cover

        if self.bind_f is not None:
            return self.bind_f(instance, obj)

        return getattr(obj, self.name)  # pragma: no cover

    def __repr__(self) -> str:
        return f"proxy {self.name}"  # pragma: no cover

    def __call__(self, instance: "LocalProxy", *args: t.Any, **kwargs: t.Any) -> t.Any:
        """Support calling unbound methods from the class. For example,
        this happens with `copy.copy`, which does
        `type(x).__copy__(x)`. `type(x)` can't be proxied, so it
        returns the proxy type and descriptor.
        """
        return self.__get__(instance, type(instance))(
            *args, **kwargs
        )  # pragma: no cover


class _ProxyIOp(_ProxyLookup):
    """Look up an augmented assignment method on a proxied object. The
    method is wrapped to return the proxy instead of the object.
    """

    __slots__ = ()

    def __init__(
        self, f: t.Optional[t.Callable] = None, fallback: t.Optional[t.Callable] = None
    ) -> None:
        super().__init__(f, fallback)

        def bind_f(instance: "LocalProxy", obj: t.Any) -> t.Callable:
            def i_op(self: t.Any, other: t.Any) -> "LocalProxy":  # pragma: no cover
                f(self, other)  # type: ignore
                return instance

            return i_op.__get__(obj, type(obj))  # pragma: no cover # type: ignore

        self.bind_f = bind_f


def _l_to_r_op(op: F) -> F:
    """Swap the argument order to turn an l-op into an r-op."""

    def r_op(obj: t.Any, other: t.Any) -> t.Any:
        return op(other, obj)  # pragma: no cover

    return t.cast(F, r_op)


class LocalProxy:
    """A proxy to the object bound to a `Local`. All operations
    on the proxy are forwarded to the bound object. If no object is
    bound, a `RuntimeError` is raised.

    ```python
    from werkzeug.local import Local
    l = Local()

    # a proxy to whatever l.user is set to
    user = l("user")

    from werkzeug.local import LocalStack
    _request_stack = LocalStack()

    # a proxy to _request_stack.top
    request = _request_stack()

    # a proxy to the session attribute of the request proxy
    session = LocalProxy(lambda: request.session)
    ```

    `__repr__` and `__class__` are forwarded, so `repr(x)` and
    `isinstance(x, cls)` will look like the proxied object. Use
    `issubclass(type(x), LocalProxy)` to check if an object is a
    proxy.

    ```python
    repr(user)  # <User admin>
    isinstance(user, User)  # True
    issubclass(type(user), LocalProxy)  # True
    ```

    Args:
        local: callable that provides the proxied object.
        name: The attribute name to look up on a `Local`. Not used if a callable is given.

    """

    __slots__ = ("__local", "__name", "__wrapped__")

    def __init__(
        self,
        local: t.Callable[[], t.Any],
        name: t.Optional[str] = None,
    ) -> None:
        object.__setattr__(self, "_LocalProxy__local", local)
        object.__setattr__(self, "_LocalProxy__name", name)

        if callable(local) and not hasattr(local, "__release_local__"):
            # "local" is a callable that is not an instance of Local or
            # LocalManager: mark it as a wrapped function.
            object.__setattr__(self, "__wrapped__", local)

    def _get_current_object(self) -> t.Any:
        """Return the current object.  This is useful if you want the real
        object behind the proxy at a time for performance reasons or because
        you want to pass the object into a different context.
        """
        if not hasattr(self.__local, "__release_local__"):  # type: ignore
            return self.__local()  # type: ignore

        try:  # pragma: no cover
            return getattr(self.__local, self.__name)  # type: ignore
        except AttributeError:  # pragma: no cover
            name = self.__name  # type: ignore
            raise RuntimeError(f"no object bound to {name}") from None

    __doc__ = _ProxyLookup(  # type: ignore
        class_value=__doc__, fallback=lambda self: type(self).__doc__, is_attr=True
    )
    # __del__ should only delete the proxy
    __repr__ = _ProxyLookup(  # type: ignore
        repr, fallback=lambda self: f"<{type(self).__name__} unbound>"
    )
    __str__ = _ProxyLookup(str)  # type: ignore
    __bytes__ = _ProxyLookup(bytes)
    __format__ = _ProxyLookup()  # type: ignore
    __lt__ = _ProxyLookup(operator.lt)
    __le__ = _ProxyLookup(operator.le)
    __eq__ = _ProxyLookup(operator.eq)  # type: ignore
    __ne__ = _ProxyLookup(operator.ne)  # type: ignore
    __gt__ = _ProxyLookup(operator.gt)
    __ge__ = _ProxyLookup(operator.ge)
    __hash__ = _ProxyLookup(hash)  # type: ignore
    __bool__ = _ProxyLookup(bool, fallback=lambda self: False)
    __getattr__ = _ProxyLookup(getattr)
    # __getattribute__ triggered through __getattr__
    __setattr__ = _ProxyLookup(setattr)  # type: ignore
    __delattr__ = _ProxyLookup(delattr)  # type: ignore
    __dir__ = _ProxyLookup(dir, fallback=lambda self: [])  # type: ignore
    # __get__ (proxying descriptor not supported)
    # __set__ (descriptor)
    # __delete__ (descriptor)
    # __set_name__ (descriptor)
    # __objclass__ (descriptor)
    # __slots__ used by proxy itself
    # __dict__ (__getattr__)
    # __weakref__ (__getattr__)
    # __init_subclass__ (proxying metaclass not supported)
    # __prepare__ (metaclass)
    __class__ = _ProxyLookup(
        fallback=lambda self: type(self), is_attr=True
    )  # type: ignore
    __instancecheck__ = _ProxyLookup(lambda self, other: isinstance(other, self))
    __subclasscheck__ = _ProxyLookup(lambda self, other: issubclass(other, self))
    # __class_getitem__ triggered through __getitem__
    __call__ = _ProxyLookup(lambda self, *args, **kwargs: self(*args, **kwargs))
    __len__ = _ProxyLookup(len)
    __length_hint__ = _ProxyLookup(operator.length_hint)
    __getitem__ = _ProxyLookup(operator.getitem)
    __setitem__ = _ProxyLookup(operator.setitem)
    __delitem__ = _ProxyLookup(operator.delitem)
    # __missing__ triggered through __getitem__
    __iter__ = _ProxyLookup(iter)
    __next__ = _ProxyLookup(next)
    __reversed__ = _ProxyLookup(reversed)
    __contains__ = _ProxyLookup(operator.contains)
    __add__ = _ProxyLookup(operator.add)
    __sub__ = _ProxyLookup(operator.sub)
    __mul__ = _ProxyLookup(operator.mul)
    __matmul__ = _ProxyLookup(operator.matmul)
    __truediv__ = _ProxyLookup(operator.truediv)
    __floordiv__ = _ProxyLookup(operator.floordiv)
    __mod__ = _ProxyLookup(operator.mod)
    __divmod__ = _ProxyLookup(divmod)
    __pow__ = _ProxyLookup(pow)
    __lshift__ = _ProxyLookup(operator.lshift)
    __rshift__ = _ProxyLookup(operator.rshift)
    __and__ = _ProxyLookup(operator.and_)
    __xor__ = _ProxyLookup(operator.xor)
    __or__ = _ProxyLookup(operator.or_)
    __radd__ = _ProxyLookup(_l_to_r_op(operator.add))
    __rsub__ = _ProxyLookup(_l_to_r_op(operator.sub))
    __rmul__ = _ProxyLookup(_l_to_r_op(operator.mul))
    __rmatmul__ = _ProxyLookup(_l_to_r_op(operator.matmul))
    __rtruediv__ = _ProxyLookup(_l_to_r_op(operator.truediv))
    __rfloordiv__ = _ProxyLookup(_l_to_r_op(operator.floordiv))
    __rmod__ = _ProxyLookup(_l_to_r_op(operator.mod))
    __rdivmod__ = _ProxyLookup(_l_to_r_op(divmod))
    __rpow__ = _ProxyLookup(_l_to_r_op(pow))
    __rlshift__ = _ProxyLookup(_l_to_r_op(operator.lshift))
    __rrshift__ = _ProxyLookup(_l_to_r_op(operator.rshift))
    __rand__ = _ProxyLookup(_l_to_r_op(operator.and_))
    __rxor__ = _ProxyLookup(_l_to_r_op(operator.xor))
    __ror__ = _ProxyLookup(_l_to_r_op(operator.or_))
    __iadd__ = _ProxyIOp(operator.iadd)
    __isub__ = _ProxyIOp(operator.isub)
    __imul__ = _ProxyIOp(operator.imul)
    __imatmul__ = _ProxyIOp(operator.imatmul)
    __itruediv__ = _ProxyIOp(operator.itruediv)
    __ifloordiv__ = _ProxyIOp(operator.ifloordiv)
    __imod__ = _ProxyIOp(operator.imod)
    __ipow__ = _ProxyIOp(operator.ipow)
    __ilshift__ = _ProxyIOp(operator.ilshift)
    __irshift__ = _ProxyIOp(operator.irshift)
    __iand__ = _ProxyIOp(operator.iand)
    __ixor__ = _ProxyIOp(operator.ixor)
    __ior__ = _ProxyIOp(operator.ior)
    __neg__ = _ProxyLookup(operator.neg)
    __pos__ = _ProxyLookup(operator.pos)
    __abs__ = _ProxyLookup(abs)
    __invert__ = _ProxyLookup(operator.invert)
    __complex__ = _ProxyLookup(complex)
    __int__ = _ProxyLookup(int)
    __float__ = _ProxyLookup(float)
    __index__ = _ProxyLookup(operator.index)
    __round__ = _ProxyLookup(round)
    __trunc__ = _ProxyLookup(math.trunc)
    __floor__ = _ProxyLookup(math.floor)
    __ceil__ = _ProxyLookup(math.ceil)
    __enter__ = _ProxyLookup()
    __exit__ = _ProxyLookup()
    __await__ = _ProxyLookup()
    __aiter__ = _ProxyLookup()
    __anext__ = _ProxyLookup()
    __aenter__ = _ProxyLookup()
    __aexit__ = _ProxyLookup()
    __copy__ = _ProxyLookup(copy.copy)
    __deepcopy__ = _ProxyLookup(copy.deepcopy)
    # __getnewargs_ex__ (pickle through proxy not supported)
    # __getnewargs__ (pickle)
    # __getstate__ (pickle)
    # __setstate__ (pickle)
    # __reduce__ (pickle)
    # __reduce_ex__ (pickle)



from fastapi import Request

from .backends.base import BaseCacheBackend


def get_request_object(kwargs: dict) -> Request:
    """
    Function to retrieve Request object in parameter if available.
    """
    request: Request = None
    for _, v in kwargs.items():
        if isinstance(v, Request):
            request = v
            break

    return request


def get_cache_backend(request: Request, cache: str = "default") -> BaseCacheBackend:
    """
    Get specific backend cache by name.

    Args:
        request: The request object.
        cache: The cache name.
    """
    cache_backend: BaseCacheBackend = request.app.state.caches[cache]
    return cache_backend



import os

from fastapi import FastAPI

app_env = os.getenv("APP_ENV", "local")

if app_env == "local":
    from .local import Settings  # noqa
elif app_env == "development":
    from .development import Settings  # noqa
elif app_env == "production":
    from .production import Settings  # noqa
else:
    raise ValueError(f"Unknown environment: {app_env}")

settings = Settings()


def init_settings(app: FastAPI) -> Settings:
    """
    Set the settings for the application.
    """

    app.state.settings = settings
    return settings



from typing import Dict

from popol.context.globals import state
from popol.context.proxy import LocalProxy

from .queue import Queue

saq_queue: Queue = LocalProxy(lambda: state.saq_queue)
saq_queues: Dict[str, Queue] = LocalProxy(lambda: state.saq_queues)



from datetime import datetime
from typing import Any, Optional, Union

from popol import json

from .base import BaseSerializer

JSON_TYPE = Union[dict, list]


class JSONSerializer(BaseSerializer):
    """
    JSON serializer.
    """

    def __init__(
        self, dumps_kwargs: Optional[dict] = None, loads_kwargs: Optional[dict] = None
    ):
        self.dumps_kwargs = dumps_kwargs or {}
        self.loads_kwargs = loads_kwargs or {}
        self.dumps_kwargs.setdefault("default", self.handler)

    def handler(self, o: Any):
        if isinstance(o, datetime):
            return str(o)
        elif hasattr(o, "serialize"):
            return o.serialize()
        raise TypeError(f"Type {type(o)} is not JSON serializable")

    def dumps(self, data: JSON_TYPE) -> str:
        """
        Serialize the data to a JSON string.
        """
        value = json.dumps(data, **self.dumps_kwargs)
        if isinstance(value, bytes):
            value = value.decode("utf-8")
        return value

    def loads(self, data: Union[str, bytes]) -> JSON_TYPE:
        """
        Deserialize a JSON string to a dictionary or list.
        """
        return json.loads(data, **self.loads_kwargs)



from abc import ABCMeta, abstractmethod
from typing import List, Sequence, TypeVar

T = TypeVar("T")


class Pagination(metaclass=ABCMeta):
    """
    Abstract class for pagination
    """

    def __init__(self, *args, **kwds) -> None:
        pass  # pragma: no cover

    @abstractmethod
    def paginate(self, data: T) -> Sequence:
        pass  # pragma: no cover

    @abstractmethod
    def get_paginated_response(self, data: T) -> dict:
        pass


class PageNumberPagination(Pagination):
    """
    Pagination class for page number
    """

    def __init__(self, page: int, page_size: int):
        self.page = page
        self.page_size = page_size

    def get_offset(self) -> int:
        return self.page_size * (self.page - 1)

    def get_limit(self) -> int:
        return self.page_size

    def paginate(self, data: Sequence) -> Sequence:
        offset = self.get_offset()
        limit = self.get_limit()
        return data[offset : offset + limit]

    def get_total_data(self, data: T) -> int:
        """
        Get total data.
        Might be useful if the data is a QuerySet or something that can compute :)
        """

        return len(data)

    def get_total_page(self, total: int, page_size: int = 10) -> List[int]:
        """
        Get total pages.
        Args:
            total: Total data.
            page_size: Page size.
        """

        if total == 0:
            return [1]  # pragma: no cover

        return list(
            range(
                1,
                (
                    total // page_size + 1
                    if total % page_size != 0
                    else total // page_size
                )
                + 1,
            )
        )

    def get_paginated_response(
        self,
        data: T,
    ) -> dict:
        """
        Return a paginated response.
        Args:
            data: Data to be paginated.
        """

        # Counting all pages
        total_data = self.get_total_data(data)
        pages = self.get_total_page(total_data, self.page_size)
        prev_page = self.page - 1
        if prev_page not in pages:
            prev_page = None  # type: ignore[assignment]

        next_page = self.page + 1
        if next_page not in pages:
            next_page = None  # type: ignore[assignment]

        # Get data per page
        data = self.paginate(data)
        rows = self.get_total_data(data)
        result = {
            "total": total_data,
            "rows": rows,
            "paging": {"next": next_page, "prev": prev_page, "pages": pages},
            "data": data,
        }
        return result



from typing import Callable

try:
    from aioredis.client import Redis
    from saq.job import Job
    from saq.queue import Queue as _Queue

except ImportError:
    errMsg = "SAQ is not installed. Please install it with `pip install saq`"
    raise RuntimeError(errMsg)

from popol import json


def _json_dumps(*args, **kwargs):
    data = json.dumps(*args, **kwargs)
    if isinstance(data, bytes):
        data = data.decode("utf-8")
    return data


class Queue(_Queue):
    @classmethod
    def from_url(
        cls,
        url: str,
        name: str = "popol",
        dump: Callable = _json_dumps,
        load: Callable = json.loads,
        **kwargs,
    ) -> "Queue":
        """Create a queue with a redis url a name."""
        return cls(url, name=name, dump=dump, load=load, **kwargs)

    def __init__(
        self,
        url: str,
        *,
        name: str = "popol",
        dump: Callable = json.dumps,
        load: Callable = json.loads,
        **kwargs,
    ):
        """
        Initialize a queue.

        Args:
            url: The redis url.
            name: The queue name.
            dump: Function to serialize job to redis.
            load: Function to deserialize job from redis.
        """

        self.url = url
        redis = Redis.from_url(url, **kwargs)
        super().__init__(redis, name, dump, load)

    async def enqueue(self, job_or_func, **kwargs) -> Job:
        """
        Enqueue a job.
        """
        return await super().enqueue(job_or_func, **kwargs)



import os
from typing import Optional

from pkg_resources import get_distribution, iter_entry_points
from typer import Argument, Context, Option, Typer, echo, prompt

from popol.utils import copy_template_dir

popol = Typer(
    name="popol", help="Popol CLI", no_args_is_help=True, invoke_without_command=True
)


def load_commands():
    """
    Load all global commands from entry point to `popol` command.
    """
    for ep in iter_entry_points("popol.commands"):
        try:
            cmd = ep.load()
            if isinstance(cmd, Typer):
                popol.add_typer(cmd)
            else:
                popol.command(ep.name, cmd)
        except Exception:
            # TODO: log error
            pass


@popol.callback()
def init(
    ctx: Context,
    version: Optional[bool] = Option(
        None, "-v", "--version", is_eager=True, help="Show version number and exit"
    ),
):
    if ctx.resilient_parsing:  # pragma: no cover
        return

    if version:
        ver = get_distribution("popol").version
        echo(f"popol v{ver}")
        ctx.exit()


@popol.command("init")
def init(outdir: str = Argument(None, help="The output directory")):
    """
    Initialize a new Popol project.
    """

    if outdir is None:
        echo("Please specify an output directory.", err=True)
        exit(1)

    project_name: str = prompt("Project name (e.g. Powerful API): ")
    project_description: str = prompt("Project description (e.g. A powerful API): ")
    echo("Initialize a new Popol project...")
    copy_template_dir(
        outdir, project_name=project_name, project_description=project_description
    )
    echo("Your project has been created in {}".format(outdir))


if "POPOL_COMMAND_INITIALIZED" not in os.environ:
    load_commands()
    os.environ["POPOL_COMMAND_INITIALIZED"] = "1"



from asyncio import iscoroutinefunction
from functools import wraps
from types import MethodType
from typing import Callable

from fastapi import Request

from ..utils import run_sync
from .backends.base import BaseCacheBackend
from .helpers import get_cache_backend, get_request_object
from .key import get_cache_key


def auto_connect(method: MethodType):
    """
    Decorator to ensure the backend cache is connected to the server.
    """
    if iscoroutinefunction(method):

        @wraps(method)
        async def wrapper(*args, **kwargs):
            self: BaseCacheBackend = args[0]
            if not self.client:
                await self.connect()
            return await method(*args, **kwargs)

    else:

        @wraps(method)
        def wrapper(*args, **kwargs):
            self: BaseCacheBackend = args[0]
            if not self.client:
                self.connect()
            return method(*args, **kwargs)

    return wrapper


def cached(
    timeout: int = 15,
    *,
    key: Callable[[Request], str] = get_cache_key,
    cache: str = "default"
) -> Callable:
    """
    Decorator to cache the result of a function.

    Args:
        timeout: Timeout in seconds.
        key: Function to generate the cache key.
        cache: The cache name.
    """

    assert callable(key), "key must be callable"

    def decorator(func: Callable) -> Callable:
        if iscoroutinefunction(func):

            @wraps(func)
            async def wrapper(*args, **kwargs):
                request = get_request_object(kwargs)
                if not request:
                    return await func(*args, **kwargs)

                cache_backend = get_cache_backend(request, cache)
                cache_key = key(request)
                cache_get = cache_backend.get
                if iscoroutinefunction(cache_get):
                    value = await cache_get(cache_key)
                else:
                    value = cache_get(cache_key)

                if value is None:
                    value = await func(*args, **kwargs)
                    cache_set = cache_backend.set
                    if iscoroutinefunction(cache_set):
                        await cache_set(cache_key, value, timeout)
                    else:
                        cache_set(cache_key, value, timeout)

                return value

        else:

            @wraps(func)
            def wrapper(*args, **kwargs):
                request = get_request_object(kwargs)
                if not request:
                    return func(*args, **kwargs)

                cache_backend = get_cache_backend(request, cache)
                cache_key = key(request)
                cache_get = cache_backend.get
                if iscoroutinefunction(cache_get):
                    value = run_sync(cache_get, cache_key)
                else:
                    value = cache_get(cache_key)

                if value is None:
                    value = func(*args, **kwargs)
                    cache_set = cache_backend.set
                    if iscoroutinefunction(cache_set):
                        run_sync(cache_set, cache_key, value, timeout)
                    else:
                        cache_set(cache_key, value, timeout)

                return value

        return wrapper

    return decorator



from fastapi import FastAPI
from starlette.datastructures import State

from .proxy import LocalProxy
from .vars import _app_ctx

current_app: FastAPI = LocalProxy(lambda: _app_ctx.get(None))
state: State = LocalProxy(lambda: current_app.state)



import asyncio

from saq.worker import Worker
from typer import Option, Typer, echo

from ...utils import get_settings, load_app
from .config import parse_config

saq = Typer(name="saq", help="SAQ Worker")


@saq.command()
def runworker(queue: str = Option(..., help="Queue to run (e.g. default)")):
    """
    Run the SAQ worker.
    """

    app = load_app()
    settings = get_settings(app)
    queue_maps, queue_settings = parse_config(settings)
    if queue not in queue_maps:
        raise RuntimeError("Queue {0!r} not found in settings".format(queue))

    q_setting = queue_settings[queue]
    queue_obj = queue_maps[queue]
    functions = q_setting.pop("functions", [])
    startup = q_setting.pop("startup", None)
    if startup and not asyncio.iscoroutinefunction(startup):
        raise RuntimeError("startup must be a coroutine")

    default_ctx = q_setting.pop("context", {})
    if not isinstance(default_ctx, dict):
        raise RuntimeError("context must be a dict")

    async def x_startup(ctx: dict):
        ctx.update(default_ctx)
        ctx["app"] = app
        ctx["state"] = app.state
        if startup:
            await startup(ctx)

    q_setting.pop("url", None)
    q_setting["startup"] = x_startup
    worker = Worker(queue_obj, functions, **q_setting)
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    echo("Starting SAQ worker... (Press CTRL-C to exit)")
    try:
        loop.run_until_complete(worker.start())
    finally:
        loop.stop()


@saq.command()
def list_queue():
    """
    List queues.
    """

    app = load_app()
    settings = get_settings(app)
    queue_maps, queue_settings = parse_config(settings)
    for q_name, queue in queue_maps.items():
        echo(f"* {q_name} ({queue.url}): {queue_settings[q_name]}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['popol',
 'popol.cache',
 'popol.cache.backends',
 'popol.cache.serializers',
 'popol.context',
 'popol.db',
 'popol.db.sqlmodel',
 'popol.email',
 'popol.jobs',
 'popol.jobs.saq',
 'popol.templates.app',
 'popol.templates.app.extensions',
 'popol.templates.app.middleware',
 'popol.templates.app.routers',
 'popol.templates.app.settings',
 'popol.templates.tests']

package_data = \
{'': ['*'], 'popol': ['templates/*']}

install_requires = \
['asyncer>=0.0.1,<0.0.2',
 'fastapi>=0.78.0,<0.79.0',
 'pydantic[email]>=1.9.1,<2.0.0',
 'typer>=0.6.1,<0.7.0',
 'uvicorn[standard]>=0.18.2,<0.19.0']

extras_require = \
{'aioredis': ['aioredis>=2.0.1,<3.0.0'],
 'aiosmtplib': ['aiosmtplib>=1.1.6,<2.0.0'],
 'all': ['redis>=4.1.4,<5.0.0',
         'saq>=0.8.0,<0.9.0',
         'sqlmodel>=0.0.6,<0.0.7',
         'pyhumps>=3.7.2,<4.0.0',
         'aioredis>=2.0.1,<3.0.0',
         'aiosmtplib>=1.1.6,<2.0.0'],
 'background_jobs': ['saq>=0.8.0,<0.9.0'],
 'cache': ['redis>=4.1.4,<5.0.0', 'aioredis>=2.0.1,<3.0.0'],
 'orm': ['sqlmodel>=0.0.6,<0.0.7', 'pyhumps>=3.7.2,<4.0.0'],
 'redis': ['redis>=4.1.4,<5.0.0'],
 'saq': ['saq>=0.8.0,<0.9.0'],
 'sqlmodel': ['sqlmodel>=0.0.6,<0.0.7', 'pyhumps>=3.7.2,<4.0.0']}

entry_points = \
{'console_scripts': ['popol = popol.__main__:popol'],
 'popol.commands': ['saq = popol.jobs.saq.cli:saq']}

setup_kwargs = {
    'name': 'popol',
    'version': '0.2.2',
    'description': 'Third party packages for your FastAPI projects',
    'long_description': '# Popol ⛅\n\n> Third party packages for your FastAPI projects ⛅\n\nPopol is a library that provides as-is tools for use on FastAPI.\n\nThis project aims to provide APIs to support your FastAPI projects without breaking existing projects. This is another version of the [Fastack](https://github.com/fastack-dev/fastack) project. Overall the available APIs are not much different from the [Fastack plugins](https://github.com/fastack-dev).\n\n## Features 🌟\n\n- [x] Project Layout\n- [x] Command Line Interface (like `flask` command)\n- [x] Pagination\n- Cache Framework\n\n    - Backends\n\n        - [x] Redis\n        - [x] Aioredis\n        - [ ] Memcached\n\n    - Serializers\n\n        - [x] JSON\n        - [x] Pickle\n        - [ ] MsgPack\n\n- ORM Integration\n\n    - [x] SQLModel (Async/Sync)\n    - [ ] Tortoise ORM\n\n- ODM Integration\n\n    - [ ] MongoEngine\n\n- [x] SMTP client (using aioredis) to send emails.\n- Background Jobs:\n\n    - [x] SAQ queue support for task scheduling\n\n\n## Installation 📚\n\n```\npip install popol\n```\n\n## Documentation 📖\n\nNot available at this time, please learn from the [examples](https://github.com/aprilahijriyan/popol/tree/main/examples).\n',
    'author': 'aprilahijriyan',
    'author_email': 'hijriyan23@gmail.com',
    'maintainer': 'aprilahijriyan',
    'maintainer_email': 'hijriyan23@gmail.com',
    'url': 'https://github.com/aprilahijriyan/popol',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from abc import ABCMeta, abstractmethod
from typing import Any, Optional


class BaseSerializer(metaclass=ABCMeta):
    """
    Base class for all serializers.
    """

    def __init__(
        self, dumps_kwargs: Optional[dict] = None, loads_kwargs: Optional[dict] = None
    ):
        pass

    @abstractmethod
    def dumps(self, data: Any) -> Any:
        """
        Serialize the data.
        """

    @abstractmethod
    def loads(self, data: Any) -> Any:
        """
        Deserialize the data.
        """



from datetime import timedelta
from typing import Any, Optional, Union

from ..decorators import auto_connect
from ..serializers.base import BaseSerializer
from .base import BaseCacheBackend

try:
    from redis.client import Pipeline, Redis
    from redis.lock import Lock

except ImportError:
    errMsg = "Redis is not installed. Please install it with `pip install redis`"
    raise RuntimeError(errMsg)


class RedisBackend(BaseCacheBackend):
    """
    Redis cache backend.
    """

    def __init__(self, serializer: BaseSerializer, **kwargs):
        """
        Initialize the Redis cache.
        """

        super().__init__(serializer, **kwargs)
        self.client: Optional[Redis] = None

    def connect(self, **kwargs) -> "RedisBackend":
        """
        Create a connection to the Redis server.
        """

        if self.client:
            return self.client

        kwargs.update(self.kwargs)
        self.client = Redis(**kwargs)
        return self

    def disconnect(self) -> None:
        """
        Disconnect from the Redis server.
        """

        if self.client:
            self.client.close()
            self.client = None

    @auto_connect
    def get(self, key: str) -> Any:
        """
        Get a value from the cache.
        """

        value = self.client.get(key)
        if value:
            return self.serializer.loads(value)
        return value

    @auto_connect
    def set(
        self, key: str, value: Any, timeout: Optional[Union[int, timedelta]] = None
    ) -> None:
        """
        Set a value in the cache.
        """

        value = self.serializer.dumps(value)
        if timeout:
            self.client.setex(key, timeout, value)
        else:
            self.client.set(key, value)

    @auto_connect
    def delete(self, key: str) -> None:
        """
        Delete a value from the cache.
        """
        self.client.delete(key)

    @auto_connect
    def clear(self) -> None:
        """
        Clear the entire cache.
        """
        self.client.flushdb()

    @auto_connect
    def incr(self, name: str, amount: int = 1) -> int:
        """
        Increment the value of an integer cached key.
        """
        return self.client.incr(name, amount)

    @auto_connect
    def decr(self, name: str, amount: int = 1) -> int:
        """
        Decrement the value of an integer cached key.
        """
        return self.client.decr(name, amount)

    @auto_connect
    def pipeline(self, **kwds) -> Pipeline:
        return self.client.pipeline(**kwds)

    @auto_connect
    def lock(self, name: str, **kwds) -> Lock:
        return self.client.lock(name, **kwds)

    def __enter__(self) -> "RedisBackend":
        """
        Context manager enter.
        """

        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Context manager exit.
        """

        self.disconnect()
